import os
import sys
import platform


MINIMUM_PYTHON_VERSION = (2,7)


class InstallerConfiguration(object):
    """
    A simple class that encapsulates all of the necessary wiring for working out paths,
    compilation flags and other details for setting up ESig.
    """
    def __init__(self, setup_location):
        """
        Constructor for the class.
        """
        self.__package_abs_root = os.path.dirname(setup_location)
    
    
    @property
    def is_x64(self):
        """
        Returns a boolean indicating whether the current system configuration is 64-bit.
        Solution obtained from Python documentation at https://docs.python.org/2/library/platform.html#platform.architecture
        """
        return sys.maxsize > 2 ** 32
    
    
    @property
    def is_windows(self):
        """
        Returns a boolean indicating whether the system is running Windows.
        True is returned if Windows is running, False otherwise.
        """
        return platform.system().lower() == 'windows'
    
    @property
    def compiler_name(self):
        """
        If running Windows, returns a string denoting the version of the Microsoft Visual C++ compiler to use for
        compiling libalgebra. If another platform is running, None is returned.
        """
        if self.is_windows:
            py_major = sys.version_info.major
            py_minor = sys.version_info.minor
        
            if py_major < 3 or (py_major == 3 and py_minor < 3):
                return 'msvc-9.0'
            else:
                if py_major == 3 and py_minor < 5:
                    return 'msvc-10.0'
                else:
                    return 'msvc-14.0'
        
        return None  # Not running Windows; returns None.
    
    
    @property
    def library_dirs(self):
        """
        
        """
        library_paths = []
        
        if self.is_windows:
            boost_path = ''
            instruction_set = 'x64' if self.is_x64 else 'win32'
            compiler_version = 'lib64' if self.is_x64 else 'lib32'
            compiler_version = '{0}-{1}'.format(compiler_version, self.compiler_name)
            
            if 'BOOST_ROOT' in os.environ:
                boost_path = os.environ['BOOST_ROOT']
            
            library_paths.append(os.path.join(boost_path, compiler_version))
            library_paths.append(os.path.join(os.path.join(boost_path,instruction_set),'lib'))
        
        return library_paths
    
    
    @property
    def include_dirs(self):
        """
        numpy is added later on.
        """
        include_dirs = [os.path.join(sys.exec_prefix, 'include')]
        
        if self.is_windows:
            include_dirs.append('.\\src\\')
            include_dirs.append('C:\\libs\\boost_1_63_0\\')
        else:
            include_dirs.append('./src/')
        
        return include_dirs
    
    @property
    def extra_compile_args(self):
        """
        Returns a list of extra compilation options, depending upon the platform specified.
        """
        args = []
        
        if self.is_windows:
            args.append('/EHsc')
            args.append('/DWINVER=0x0601')
            args.append('/D_WIN32_WINNT=0x0601')
            args.append('/D_SCL_SECURE_NO_WARNINGS')
            args.append('/bigobj')
        else:
            args.append('-Wno-unused-but-set-variable')
        
        return args
    
    
    @property
    def linker_args(self):
        """
        Returns a list of args used to be used by the linker when compiling.
        Options are platform-specific.
        """
        args = []
        
        if not self.is_windows:
            args.append('-static')
        
        return args
    
    
    @property
    def boost_libraries(self):
        """
        Returns a list of Boost libraries that are used by libalgebra.
        Note that no libraries are returned if Windows is running - the V C++ compiler should pick the libraries automatically.
        """
        libraries = []
        
        if not self.is_windows:
            libraries.append('boost_system')
            libraries.append('boost_thread')
        
        return libraries
    
    
    def check_python_version(self):
        """
        Runs a simple test -- is the version of Python running the installer acceptable?
        If not, bomb out with a message.
        The minimum version is obtained from the tuple MINIMUM_PYTHON_VERSION.
        """
        if sys.version_info < MINIMUM_PYTHON_VERSION:
            sys.exit("Python 2.7.x or greater is required for esig. Consider upgrading your Python environment.")
    

    def get_version(self):
        """
        Returns a string representing the version number of ESig.
        The version is extracted from the VERSION file, found in the base of the package.
        """
        version_path = os.path.join(self.__package_abs_root, 'VERSION')
        
        with open(version_path, 'r', encoding='utf-8') as version_file:
            version = (version_file.read().strip()).replace(' ', '.')
        
        return version
    
    
    def get_long_description(self):
        """
        Returns a string representing the long description of ESig.
        This is extracted from the README.md file in the base of the package.
        """
        readme_path = os.path.join(self.__package_abs_root, 'README.md')
        
        with open(readme_path, 'r', encoding='utf-8') as f:
            long_description = f.read()
        
        return long_description


import numpy as np
from esig import tosig as ts
import auxiliaryfunct as ax
import unittest
import esigtests as rado
from unittest import TestCase

# Tree-like equivalence of paths and equivalence class signature invariance. Non-generic paths used since pruning is generically unnecessary.

a=[[0.0,0],[1,0],[1,1],[1,0],[2,0],[3,1],[2,2],[1,1],[2,2],[1,3],[2,2],[3,3],[2,2],[3,1],[4,1],[3,1],[2,0],[2,-1],[2,0],[1,0],[1,-1],[1,0],[0,0]]

b=[[0.0,0], [1,3], [0,0], [1,5], [2,5], [1,5], [0,6], [1,5], [0,0]]

c=[[0.0,0], [1,1], [3,1], [2,1], [1,1], [2,0]]

aa=ax.fullprune(a)
bb=ax.fullprune(b)
cc=ax.fullprune(c)
a=np.array(a)
b=np.array(b)
c=np.array(c)
aa=np.array(aa)
bb=np.array(bb)
cc=np.array(cc)
sigdiffa= ts.stream2sig(a,3)-ts.stream2sig(aa,3)
sigdiffb = ts.stream2sig(b,3)-ts.stream2sig(bb,3)
sigdiffc = ts.stream2sig(c,3)-ts.stream2sig(cc,3)


y=ax.random_path(1,range(-10,11),3)

x=[[0.0,0,0]]
xx = x

xx.append(y[0])
x=np.array(xx)

path1=ax.random_path(5,range(-10,11),3)
path2=ax.random_path(5,range(-10,11),3)
#print(path1)
#print(ax.concatenate(path1,path2))

sig1=ax.tensoraslevels(ts.stream2sig(np.array(path1),2),3,2)
#print(sig1)
sig2=ax.tensoraslevels(ts.stream2sig(np.array(path2),2),3,2)
#print(sig2)
sig_concat=ax.tensoraslevels(ts.stream2sig(ax.concatenate(path1,path2),2),3,2)
chen=ax.chen(sig1,sig2)
#print(chen)
#print(sig_concat)


logsig=ax.exponentiateTensor( ax.tensoraslevels( ax.Logsigastensor( np.array(path1),3),3,3))



#c=ax.Logsigastensor(a,3)
#b=ts.stream2sig(a,3)
#d=ax.exponentiateTensor(ax.tensoraslevels(c,2,3))
#print(b)
#print(d)
#b=np.array(b)
#sigdiff= ts.stream2sig(a,2)-ts.stream2sig(b,2)
#print(sigdiff)

#if all(sigdiff)==0:
 #   print("ok")


            
# Test Chen's theorem using tensor multiplication
# Compute esig of linear functions by hand and see it matches

# Exponentiate logsig as a tensor and see it matches esig 

 
#a=[1,2]
#sig1=ax.linsig(a,3)
#b=[[0.0,0], [1,2]]
#b=np.array(b)
#sig2=ts.stream2sig(np.array(b),3)
#print(sig1-sig2)


#print(ax.fullunbracket(string))
#a=ax.Lietotensor(string)
#print(a)



#print(expand)

#L=ax.Lietotensor(string)
#print(L)
#print(ax.Lietotensor('1'))

#print(ax.changeofbasis(2,3))


#print(ax.Lietotensor(string))
#tensor_product=ax.fullunbracket(string)
#copy= tensor_product[1:-1]
#print(copy)

#term=ax.spotfirstterm(copy)
#print(type(term))
#print(term)
#d=ax.distribute(term)
#print(d)
#print(type(d))
##print(copy.replace(term,d))
#copy=copy.replace(term,ax.distribute(term))
#print(copy)
#term=ax.spotfirstterm(copy)
#print(term)
#print(type(term))
#a=ax.fullunbracket(string)

#print(a)
#print(len(a))
#b=a[1:-1]
#c=ax.spotterm(b)[1][1]
#print(c)

#term='1x2x(2x1-1x2)x3'
#print(ax.distribute(term))

class TestESIG(TestCase):

    def test_equivalence(self):
        self.assertEqual(all(sigdiffa),0)
        self.assertEqual(all(sigdiffb),0)
        self.assertEqual(all(sigdiffc),0)

    def test_linsig(self):
        self.assertEqual(all(ax.linsig(y[0],4) - ts.stream2sig(x,4)),0)

    def test_logsig(self):
        self.assertEqual(all(ts.stream2sig(np.array(path1),3)[1:]- logsig),0)

    def test_chen(self):
        self.assertEqual(all(np.concatenate((np.array(chen)-np.array(sig_concat)).flatten())),0)

    def test_compare_reverse(self):
        self.assertEqual(rado.compare_reverse_test(100,[-1,0,1],2,4),'ok')

    def test_shuffle_test(self):
        self.assertEqual(rado.shuffle_test(100,[-1,0,1],3,3),'ok')
    
    def test_reparametrizing(self):
        self.assertEqual(rado.reparametrizing_test(10,[-1,0,1],2,3,3),'ok')


import numpy as np
import esig.tosig as ts
import math
import pandas as pd
import random
import itertools
import auxiliaryfunct as ax

"""
This file contains unit tests of the esig library as well as auxiliary functions needed to make the tests run.

The tests are run each time esig is installed, to make sure the installation works correctly.

Separate descriptions are available for each function

"""



def compare_reverse_test(l,a,dims,degs):
    """ Compare signature of a random path with its inverse
    
    If a path has steps [a,b,c], its inverse has steps [c,b,a]

    Args:
        l (int): length of the random path to be produced
        a (list): list of modifiers applied at each step of the random path
        dims (int): number of dimensions each path step has
        degs (int): number of signature degrees to be used in the test

    Returns:
        str: "ok" if the test has been successful

    Raises:
        ValueError: If the signature of the inverse path is not actually computed for the inverse of the original path
    
    Example:
        >>> compare_reverse_test(100,[-1,0,1],2,4)
        "ok"
    """

    input = np.array(ax.random_path(l,a,dims))

    # make a reverse path
    rev = list(reversed(input))
    rev = [list(x) for x in rev]
    rev = np.array(rev)
    out_sig = ts.stream2sig(input,degs)
    out_rev = ts.stream2sig(rev,degs)

    # calculate corresponding 'k' addresses for the signature elements of the original and the inverse paths
    pos = ts.sigkeys(dims,degs).split()
    how_long = []
    pos_out = []
    pos_rev = []
    for row in pos:
        row = row.strip('(')
        row = row.strip(')')
        row = row.split(',')
        pos_out.append(row)
        pos_rev.append(list(reversed(row)))
        how_long.append(len(row))

    tmp = []
    for x in how_long:
        if x % 2 == 0: # even numbers
            tmp.append(-1)
        else: # odd numbers
            tmp.append(1)
    how_long = tmp

    # match 'k' addresses for the signature elements of the original and the inverse paths
    pos_out = [str(x) for x in pos_out]
    pos_rev = [str(x) for x in pos_rev]
    tmp_out = list(zip(pos_out, how_long, out_sig))
    tmp_rev = list(zip(pos_rev, how_long, out_rev))
    tmp_out = pd.DataFrame(tmp_out)
    tmp_rev = pd.DataFrame(tmp_rev)
    out_and_rev = pd.merge(tmp_out,tmp_rev, left_on=0, right_on=0)

    # check if the differences are correct (corresponding addresses of odd length have opposite signs, corresponding addresses with even length have the sae signs)

    out = out_and_rev['1_x']*out_and_rev['2_x'] + out_and_rev['2_y']
    if out[0] != 2:
        raise ValueError
    for row in out[1:]:
        if row >= -0.00001 and row <= 0.00001:
            pass
        else:
            raise ValueError
    return "ok"


def shuffle_test (l,a,dims,degs):

    """ Evaluates whether a product sum of shuffles is equal to the multiplication evaluated pairs of signature elements
    
    stream2sig and stream2logsig functions included in the esig library return a list of values according to (a) the dimensions present in the data and (b) signature degrees.
    sigkeys and logsigkeys functions included in the esig library return keys which correspond to each of the signature element produced with either stream2sig and stream2logsig functions.
    
    Properties of path signature are such that for any 2 keys, if one multiplies the values which they correspond to, the result of the multiplication is equal to a sum of 
    values for all keys which are shuffles of the two keys. This function evaluates whether this property of path signature holds true for all possible pairs of
    signature elements whose keys, together, are shorter or as long as the longest keys of values included in the signature.
    
    Args:
        l (int): length of a random path to be produced
        a (list): list of modifiers applied at each step of the random path
        dims (int): number of dimensions each path step has
        degs (int): number of signature degrees to be used in the test

    Returns:
        str: "ok" if the test has been successful

    Raises:
        ValueError: If the sum of values corresponding to key shuffles is not equal to the multiplication of the pair of values with corresponding keys

    Example:
        >>> shuffle_test(100,[-1,0,1],3,3)
        "ok"

    """

    test_path = np.array(ax.random_path(l,a,dims)) # generates a random path given number of path steps 'l', possible moves at each step 'a', 
                                                # and the number dimensions at each path step 'dims'
    signature = list(ts.stream2sig(test_path,degs)) # generates signature for the provided random path and chosen number of signature degrees 'degs'

    #5 LINES BELOW COULD BE A SEPARATE TEST
    if signature[0] != 1: # check if the first signature element is 1
        print("The first signature element has a value different from 0 in shuffle_test")
        raise ValueError
    else:
        pass

    keys = ts.sigkeys(dims,degs).split() # generates keys of the signature for the provided random path
    keys2 = []
    lengths = []
    for row in keys:
        row = row.strip('(')
        row = row.strip(')')
        if len(row) > 0:
            row = row.split(',')
            lengths.append(len(row))
            keys2.append(row)
        else:
            pass

    del signature[0]

    keys3 = list(zip(keys2,lengths,signature))




    max_mutant_length = len(keys2[-1])

    # obtain all possible mutations which are no longer than the max_mutant_length
    mutants = []
    mutant_parent_indices = []
    for row_ind, row in enumerate(keys3):
        vals = []
        inds = []
        for col_ind, col in enumerate(keys3):
            if row_ind != col_ind:
                tmp = list(row[0]) + list(col[0])
                if len(tmp) <= max_mutant_length:
                    val = [list(row[0]), list(col[0])]
                    vals.append(val)
                    ind = [row_ind,col_ind]
                    inds.append(ind)
        mutants.extend(vals)
        mutant_parent_indices.extend(inds)


    # generate shuffles, get results
    shuffle_prep = list(zip(mutants,mutant_parent_indices))
    for row in shuffle_prep:
        my_shuffles = ax.shuffles(row[0][0],row[0][1])

        # sum shuffles, knowing their values in keys3 table
        shuffle_product = 0
        for shuffle in my_shuffles:
            for x in keys3:
                if str(x[0]) != str(shuffle):
                    pass
                else:
                    shuffle_product = shuffle_product + x[2]
         
        # multiply components, knowing their indics in keys3 table
        parent1 = row[1][0]
        parent1 = keys3[parent1][2]
        parent2 = row[1][1]
        parent2 = keys3[parent2][2]
        mutant_parent_multiplication = parent1*parent2

        # shuffle product - multiplication product into 'solutions' variable
        solution = shuffle_product-mutant_parent_multiplication
        if solution >= -0.00001 and solution <= 0.00001:
            pass
        else:
            raise ValueError
    return 'ok'



# reparametrizing test
def reparametrizing_test(l,a,dims,degs,reparam_factor):
    """ Checks if the path signature for a reparametrized path is identical to the original signature. 
    
    The test works for any path which is linear everywhere except of a list of inflexion points. 
    New steps are added which fall on the lines connecting the steps of the inputted path. The overall shape of the path is the same as a result


    Args:
        l (int): length of a random path to be produced
        a (list): list of modifiers applied at each step of the random path
        dims (int): number of dimensions each path step has
        degs (int): number of signature degrees to be used in the test
        reparam_factor (int): factor by which the number of steps is increased along the existing path

    Returns:
        str: "ok" if the test has been successful
        
    Raises:
        ValueError: If reparametrized path put into stream2sig function resulted in a different path

    Examples:
        >>> print (reparametrizing_test(10,[-1,0,1],2,3,3))
        "ok"

    """
    path_orig = ax.random_path(l,a,dims) # generates a random path given number of path steps 'l', possible moves at each step 'a', 
                                                    # and the number dimensions at each path step 'dims'
    mid_step = 1.0/reparam_factor
    # obtaining additional mid-steps
    path_reparam = []
    for ind, step in enumerate(path_orig):
        path_reparam.append(step)
        next_step = ind + 1
        change = []
        if next_step < len(path_orig):
            change = [x1 - x2 for (x1, x2) in zip(path_orig[next_step], step)]
            for x in range(1,reparam_factor):
                mid_step_vals = np.array(change)*mid_step*x + step
                path_reparam.append(list(mid_step_vals))

    orig_signature = list(ts.stream2sig(np.array(path_orig), degs))
    reparam_signature = list(ts.stream2sig(np.array(path_reparam), degs))
    tmp = zip(orig_signature,reparam_signature)

    solutions = [x1 - x2 for (x1, x2) in tmp]
    for solution in solutions:
        if solution >= -0.00001 and solution <= 0.00001:
            pass
        else:
            raise ValueError
    return 'ok'








import os
import esig
import unittest


# Two global variables, providing information on where the test modules are to be found,
# and the naming convention for the modules.
TEST_MODULE_ROOT = os.path.join(esig.ESIG_PACKAGE_ROOT, 'tests')
TEST_MODULE_PATTERN = 'test*.py'


def get_suite():
    """
    Constructs and returns a Python unittest suite object.
    This can be used to run the unit tests for ESig.
    
    Args:
        None
    
    Returns:
        suite: A Python unittest suite, referring to all tests specified within the tests package.
    
    """
    loader = unittest.TestLoader()
    
    # The test suite is constructed from all modules matching the filename pattern specified by TEST_MODULE_PATTERN.
    # All modules are to be found in TEST_MODULE_ROOT.
    suite = loader.discover(TEST_MODULE_ROOT, pattern=TEST_MODULE_PATTERN)
    return suite


def run_tests(output_path=None):
    """
    Acquires the ESig test suite, and runs the tests.
    The test log can be optionally saved to a file; provide a string to the file you wish to save to as output_path.
    
    Args:
        output_path: Optional. A string representing the path to which the output log should be saved.
    
    Returns:
        None
    """
    f = None

    if output_path is not None:
        f = open(output_path, 'w')
    
    esig_test_suite = get_suite()

    runner = unittest.TextTestRunner(f)
    runner.run(esig_test_suite)
    
    if output_path is not None:
        f.close()


#
# Module containing an extension of the setuptools build_ext class.
# Part of the esig Python package.
# 
# Author: David Maxwell <maxwelld90@gmail.com>
# Date: 2017-07-07
#

__author__ = 'David Maxwell <dmaxwell@turing.ac.uk>'
__date__ = '2017-07-07'

from setuptools.command.build_ext import build_ext

class NumpyExtensionCommand(build_ext):
    """
    Extends the build_ext class.
    Overrides the run() method.
    """
    def run(self):
        """
        Attempts to import numpy and append the result of numpy.get_include() to the self.include_dirs list.
        Why is this required?
        
        Args:
            self (NumpyExtensionCommand): Instance of self.
        
        Returns:
            None
        """
        import numpy
        self.include_dirs.append(numpy.get_include())
        build_ext.run(self)


#!/usr/bin/env python

#
# The ESig Python package
# Basic package functions
#


import os


__author__ = 'David Maxwell <dmaxwell@turing.ac.uk>'
__date__ = '2017-07-21'


ESIG_PACKAGE_ROOT = os.path.abspath('.')


def get_version():
    """
    Returns the version number of the ESig package.
    The version is obtained from the VERSION file in the root of the package.
    
    Args:
        None
    
    Returns:
        string: The package version number. In format 'major.minor.release'.
    """
    f = open('../VERSION', 'r')
    version_string = f.read().split(' ')
    f.close()
    
    return '.'.join(version_string)

def is_library_loaded():
    """
    Determines whether the tosig Python extension can be successfully loaded.
    If the library cannot be loaded successfully, debugging information can be obtained from get_library_load_error().
    
    Args:
        None
    Returns:
        boolean: True iif the library can be loaded successfully; False otherwise.
    """
    try:
        from esig import tosig
    except ImportError:
        return False
    
    return True


def get_library_load_error():
    """
    Returns a string containing the message of the exception raised when attempting to import tosig.
    If no exception is raised when attempting to import, None is returned.
    
    Args:
        None
    Returns:
        string: The message associated with the exception when attempting to import.
        None: If no exception is raised when importing, None is returned.
    """
    try:
        from esig import tosig
        return None
    except ImportError as e:
        return e.msg
        


import os
from setuptools import setup, find_packages, Extension
from esig.setuptools_helpers import configuration, distribution, extensions


__author__ = 'David Maxwell'
__date__ = '2017-07-11'


configuration = configuration.InstallerConfiguration(os.path.realpath(__file__))
configuration.check_python_version()


esig_extension = Extension(
    'esig.tosig',
    sources=['src/C_tosig.c', 'src/Cpp_ToSig.cpp', 'src/ToSig.cpp'],
    depends=['src/ToSig.h', 'src/C_tosig.h', 'src/ToSig.cpp', 'src/switch.h'],
    include_dirs=configuration.include_dirs,  # Boost include directory must be added when building the extension.
    library_dirs=configuration.library_dirs,  # Boost compiled libraries directory must be added when building the extension.
    libraries=configuration.boost_libraries,
    extra_compile_args=configuration.extra_compile_args,
    extra_link_args=configuration.linker_args,  # This may need to be tweaked to get it to work on Linux.
)


setup(
    name='esig',
    version=configuration.get_version(),
    
    author='Terry Lyons',
    author_email='software@lyonstech.net',
    url='https://lyonstech.net/software',
    license='GPLv3',

    keywords='data streams rough paths signatures', 
    
    description="This package provides \"rough path\" tools for analysing vector time series.",
    long_description=configuration.get_long_description(),
    
    include_package_data=True,
    packages=find_packages(),  # Used for bdist_wheel.
    test_suite='esig.tests.get_suite',
    
    distclass=distribution.BinaryDistribution,
    ext_modules=[esig_extension],
    
    install_requires=['numpy==1.12.0'],
    setup_requires=['numpy==1.12.0'],
    tests_require=['numpy>=1.12.0'],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Mathematics',
        ],
    
    cmdclass={
        'build_ext': extensions.NumpyExtensionCommand,
    },
)